"""Pre-processing module."""

__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2019-11-25'

import logging
import os
import traceback
from typing import Tuple
import multiprocessing

import numpy as np
import requests
import psutil

import pandas as pd
import xarray as xr
from matplotlib import pyplot as plt

from mlair.data_handler import DataCollection, AbstractDataHandler
from mlair.helpers import TimeTracking, to_list, tables
from mlair.configuration import path_config
from mlair.helpers.join import EmptyQueryResult
from mlair.run_modules.run_environment import RunEnvironment


class PreProcessing(RunEnvironment):
    """
    Pre-process your data by using this class.

    Schedule of pre-processing:
        #. load and check valid stations (either download or load from disk)
        #. split subsets (train, val, test, train & val)
        #. create small report on data metrics

    Required objects [scope] from data store:
        * all elements from `DEFAULT_ARGS_LIST` in scope preprocessing for general data loading
        * all elements from `DEFAULT_ARGS_LIST` in scopes [train, val, test, train_val] for custom subset settings
        * `fraction_of_training` [.]
        * `experiment_path` [.]
        * `use_all_stations_on_all_data_sets` [.]

    Optional objects
        * all elements from `DEFAULT_KWARGS_LIST` in scope preprocessing for general data loading
        * all elements from `DEFAULT_KWARGS_LIST` in scopes [train, val, test, train_val] for custom subset settings

    Sets
        * `stations` in [., train, val, test, train_val]
        * `generator` in [train, val, test, train_val]
        * `transformation` [.]

    Creates
        * all input and output data in `data_path`
        * latex reports in `experiment_path/latex_report`

    """

    def __init__(self):
        """Set up and run pre-processing."""
        super().__init__()
        self._run()

    def _run(self):
        stations = self.data_store.get("stations")
        data_handler = self.data_store.get("data_handler")
        _, valid_stations = self.validate_station(data_handler, stations,
                                                  "preprocessing")  # , store_processed_data=False)
        if len(valid_stations) == 0:
            raise ValueError("Couldn't find any valid data according to given parameters. Abort experiment run.")
        self.data_store.set("stations", valid_stations)
        self.split_train_val_test()
        if self.data_store.get('oversampling_method')=='bin_oversampling':
            logging.debug("Apply Oversampling")
            self.apply_oversampling()
        else:
            logging.debug("No Oversampling")
        self.report_pre_processing()
        self.prepare_competitors()

    def apply_oversampling(self):
        data = self.data_store.get('data_collection', 'train')
        bins = self.data_store.get('oversampling_bins')
        rates_cap = self.data_store.get('oversampling_rates_cap')
        histogram = np.array(bins)
        #get min and max of the whole data
        total_min = 0
        total_max = 0
        for station in data:
            total_min = np.minimum(np.amin(station.get_Y(as_numpy=True)), total_min)
            total_max = np.maximum(np.amax(station.get_Y(as_numpy=True)), total_max)
        bin_edges = []
        for station in data:
            # Create histogram for each station
            hist, bin_edges = np.histogram(station.get_Y(as_numpy=True), bins=bins, range=(total_min, total_max))
            # Add up histograms
            histogram = histogram + hist
        # Scale down to most frequent class=1
        histogram = 1/np.amax(histogram) * histogram
        # Get Oversampling rates (with and without cap)
        oversampling_rates = 1 / histogram
        oversampling_rates_capped = np.minimum(oversampling_rates, rates_cap)
        # Add to datastore
        self.data_store.set('oversampling_rates', oversampling_rates, 'train')
        self.data_store.set('oversampling_rates_capped', oversampling_rates_capped, 'train')
        self.data_store.set('oversampling_bin_edges', bin_edges)
        #Y = None
        #Y_extreme = None
        for station in data:
            station.apply_oversampling(bin_edges, oversampling_rates_capped)
            '''
            if Y is None:
                Y = station._Y
                Y_extreme = station._Y_extreme
            else:
                Y = xr.concat([Y, station._Y], dim="Stations")
                Y_extreme = xr.concat([Y_extreme, station._Y_extreme], dim="Stations")
        self.data_store.set('Oversampling_Y', Y)
        self.data_store.set('Oversampling_Y_extreme', Y_extreme)
        '''


    def report_pre_processing(self):
        """Log some metrics on data and create latex report."""
        logging.debug(20 * '##')
        n_train = len(self.data_store.get('data_collection', 'train'))
        n_val = len(self.data_store.get('data_collection', 'val'))
        n_test = len(self.data_store.get('data_collection', 'test'))
        n_total = n_train + n_val + n_test
        logging.debug(f"Number of all stations: {n_total}")
        logging.debug(f"Number of training stations: {n_train}")
        logging.debug(f"Number of val stations: {n_val}")
        logging.debug(f"Number of test stations: {n_test}")
        self.create_latex_report()

    def create_latex_report(self):
        """
        Create tables with information on the station meta data and a summary on subset sample sizes.

        * station_sample_size.md: see table below as markdown
        * station_sample_size.tex: same as table below as latex table
        * station_sample_size_short.tex: reduced size table without any meta data besides station ID, as latex table

        All tables are stored inside experiment_path inside the folder latex_report. The table format (e.g. which meta
        data is highlighted) is currently hardcoded to have a stable table style. If further styles are needed, it is
        better to add an additional style than modifying the existing table styles.

        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | stat. ID   | station_name                              |   station_lon |   station_lat |   station_alt |   train |   val |   test |
        +============+===========================================+===============+===============+===============+=========+=======+========+
        | DEBW013    | Stuttgart Bad Cannstatt                   |        9.2297 |       48.8088 |           235 |    1434 |   712 |   1080 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | DEBW076    | Baden-Baden                               |        8.2202 |       48.7731 |           148 |    3037 |   722 |    710 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | DEBW087    | Schwäbische_Alb                           |        9.2076 |       48.3458 |           798 |    3044 |   714 |   1087 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | DEBW107    | Tübingen                                  |        9.0512 |       48.5077 |           325 |    1803 |   715 |   1087 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | DEBY081    | Garmisch-Partenkirchen/Kreuzeckbahnstraße |       11.0631 |       47.4764 |           735 |    2935 |   525 |    714 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | # Stations | nan                                       |      nan      |      nan      |           nan |       6 |     6 |      6 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+
        | # Samples  | nan                                       |      nan      |      nan      |           nan |   12253 |  3388 |   4678 |
        +------------+-------------------------------------------+---------------+---------------+---------------+---------+-------+--------+

        """
        meta_data = ['station_name', 'station_lon', 'station_lat', 'station_alt']
        meta_round = ["station_lon", "station_lat", "station_alt"]
        precision = 4
        path = os.path.join(self.data_store.get("experiment_path"), "latex_report")
        path_config.check_path_and_create(path)
        names_of_set = ["train", "val", "test"]
        df = self.create_info_df(meta_data, meta_round, names_of_set, precision)
        column_format = tables.create_column_format_for_tex(df)
        tables.save_to_tex(path=path, filename="station_sample_size.tex", column_format=column_format, df=df)
        tables.save_to_md(path=path, filename="station_sample_size.md", df=df)
        df_nometa = df.drop(meta_data, axis=1)
        column_format = tables.create_column_format_for_tex(df)
        tables.save_to_tex(path=path, filename="station_sample_size_short.tex", column_format=column_format,
                           df=df_nometa)
        tables.save_to_md(path=path, filename="station_sample_size_short.md", df=df_nometa)
        # df_nometa.to_latex(os.path.join(path, "station_sample_size_short.tex"), na_rep='---',
        #                    column_format=column_format)
        df_descr = self.create_describe_df(df_nometa)
        column_format = tables.create_column_format_for_tex(df_descr)
        tables.save_to_tex(path=path, filename="station_describe_short.tex", column_format=column_format, df=df_descr)
        tables.save_to_md(path=path, filename="station_describe_short.md", df=df_descr)
        # df_descr.to_latex(os.path.join(path, "station_describe_short.tex"), na_rep='---', column_format=column_format)

    @staticmethod
    def create_describe_df(df, percentiles=None, ignore_last_lines: int = 2):
        if percentiles is None:
            percentiles = [.05, .1, .25, .5, .75, .9, .95]
        df_descr = df.iloc[:-ignore_last_lines].astype('float32').describe(
            percentiles=percentiles).astype("int32", errors="ignore")
        df_descr = pd.concat([df.loc[['# Samples']], df_descr]).T
        df_descr.rename(columns={"# Samples": "no. samples", "count": "no. stations"}, inplace=True)
        df_descr_colnames = list(df_descr.columns)
        df_descr_colnames = [df_descr_colnames[1]] + [df_descr_colnames[0]] + df_descr_colnames[2:]
        df_descr = df_descr[df_descr_colnames]
        return df_descr

    def create_info_df(self, meta_data, meta_round, names_of_set, precision):
        df = pd.DataFrame(columns=meta_data + names_of_set)
        for set_name in names_of_set:
            data = self.data_store.get("data_collection", set_name)
            for station in data:
                station_name = str(station.id_class)
                df.loc[station_name, set_name] = station.get_Y()[0].shape[0]
                if df.loc[station_name, meta_data].isnull().any():
                    df.loc[station_name, meta_data] = station.id_class.meta.loc[meta_data].values.flatten()
            df.loc["# Samples", set_name] = df.loc[:, set_name].sum()
            assert len(data) == df.loc[:, set_name].count() - 1
            df.loc["# Stations", set_name] = len(data)
        df[meta_round] = df[meta_round].astype(float).round(precision)
        df.sort_index(inplace=True)
        df = df.reindex(df.index.drop(["# Stations", "# Samples"]).to_list() + ["# Stations", "# Samples"], )
        df.index.name = 'stat. ID'
        return df

    def split_train_val_test(self) -> None:
        """
        Split data into subsets.

        Currently: train, val, test and train_val (actually this is only the merge of train and val, but as an separate
        data_collection). IMPORTANT: Do not change to order of the execution of create_set_split. The train subset needs
        always to be executed at first, to set a proper transformation.
        """
        fraction_of_training = self.data_store.get("fraction_of_training")
        stations = self.data_store.get("stations")
        train_index, val_index, test_index, train_val_index = self.split_set_indices(len(stations),
                                                                                     fraction_of_training)
        subset_names = ["train", "val", "test", "train_val"]
        if subset_names[0] != "train":  # pragma: no cover
            raise AssertionError(f"Make sure, that the train subset is always at first execution position! Given subset"
                                 f"order was: {subset_names}.")
        for (ind, scope) in zip([train_index, val_index, test_index, train_val_index], subset_names):
            self.create_set_split(ind, scope)

    @staticmethod
    def split_set_indices(total_length: int, fraction: float) -> Tuple[slice, slice, slice, slice]:
        """
        Create the training, validation and test subset slice indices for given total_length.

        The test data consists on (1-fraction) of total_length (fraction*len:end). Train and validation data therefore
        are made from fraction of total_length (0:fraction*len). Train and validation data is split by the factor 0.8
        for train and 0.2 for validation. In addition, split_set_indices returns also the combination of training and
        validation subset.

        :param total_length: list with all objects to split
        :param fraction: ratio between test and union of train/val data

        :return: slices for each subset in the order: train, val, test, train_val
        """
        pos_test_split = int(total_length * fraction)
        train_index = slice(0, int(pos_test_split * 0.8))
        val_index = slice(int(pos_test_split * 0.8), pos_test_split)
        test_index = slice(pos_test_split, total_length)
        train_val_index = slice(0, pos_test_split)
        return train_index, val_index, test_index, train_val_index

    def create_set_split(self, index_list: slice, set_name: str) -> None:
        # get set stations
        stations = self.data_store.get("stations", scope=set_name)
        if self.data_store.get("use_all_stations_on_all_data_sets"):
            set_stations = stations
        else:
            set_stations = stations[index_list]
        logging.debug(f"{set_name.capitalize()} stations (len={len(set_stations)}): {set_stations}")
        # create set data_collection and store
        data_handler = self.data_store.get("data_handler")
        collection, valid_stations = self.validate_station(data_handler, set_stations, set_name)
        self.data_store.set("stations", valid_stations, scope=set_name)
        self.data_store.set("data_collection", collection, scope=set_name)

    def validate_station(self, data_handler: AbstractDataHandler, set_stations, set_name=None,
                         store_processed_data=True):
        """
        Check if all given stations in `all_stations` are valid.

        Valid means, that there is data available for the given time range (is included in `kwargs`). The shape and the
        loading time are logged in debug mode.

        :return: Corrected list containing only valid station IDs.
        """
        t_outer = TimeTracking()
        logging.info(f"check valid stations started{' (%s)' % (set_name if set_name is not None else 'all')}")
        # calculate transformation using train data
        if set_name == "train":
            logging.info("setup transformation using train data exclusively")
            self.transformation(data_handler, set_stations)
        # start station check
        collection = DataCollection(name=set_name)
        valid_stations = []
        kwargs = self.data_store.create_args_dict(data_handler.requirements(), scope=set_name)
        use_multiprocessing = self.data_store.get("use_multiprocessing")

        if multiprocessing.cpu_count() > 1 and use_multiprocessing:  # parallel solution
            logging.info("use parallel validate station approach")
            pool = multiprocessing.Pool(
                min([psutil.cpu_count(logical=False), len(set_stations), 16]))  # use only physical cpus
            logging.info(f"running {getattr(pool, '_processes')} processes in parallel")
            output = [
                pool.apply_async(f_proc, args=(data_handler, station, set_name, store_processed_data), kwds=kwargs)
                for station in set_stations]
            for i, p in enumerate(output):
                dh, s = p.get()
                logging.info(f"...finished: {s} ({int((i + 1.) / len(output) * 100)}%)")
                if dh is not None:
                    collection.add(dh)
                    valid_stations.append(s)
            pool.close()
        else:  # serial solution
            logging.info("use serial validate station approach")
            for station in set_stations:
                dh, s = f_proc(data_handler, station, set_name, store_processed_data, **kwargs)
                if dh is not None:
                    collection.add(dh)
                    valid_stations.append(s)

        logging.info(f"run for {t_outer} to check {len(set_stations)} station(s). Found {len(collection)}/"
                     f"{len(set_stations)} valid stations.")
        return collection, valid_stations

    def validate_station_old(self, data_handler: AbstractDataHandler, set_stations, set_name=None,
                             store_processed_data=True):
        """
        Check if all given stations in `all_stations` are valid.

        Valid means, that there is data available for the given time range (is included in `kwargs`). The shape and the
        loading time are logged in debug mode.

        :return: Corrected list containing only valid station IDs.
        """
        t_outer = TimeTracking()
        logging.info(f"check valid stations started{' (%s)' % (set_name if set_name is not None else 'all')}")
        # calculate transformation using train data
        if set_name == "train":
            logging.info("setup transformation using train data exclusively")
            self.transformation(data_handler, set_stations)
        # start station check
        collection = DataCollection()
        valid_stations = []
        kwargs = self.data_store.create_args_dict(data_handler.requirements(), scope=set_name)
        for station in set_stations:
            try:
                dp = data_handler.build(station, name_affix=set_name, store_processed_data=store_processed_data,
                                        **kwargs)
                collection.add(dp)
                valid_stations.append(station)
            except (AttributeError, EmptyQueryResult):
                continue
        logging.info(f"run for {t_outer} to check {len(set_stations)} station(s). Found {len(collection)}/"
                     f"{len(set_stations)} valid stations.")
        return collection, valid_stations

    def transformation(self, data_handler: AbstractDataHandler, stations):
        if hasattr(data_handler, "transformation"):
            kwargs = self.data_store.create_args_dict(data_handler.requirements(), scope="train")
            transformation_dict = data_handler.transformation(stations, **kwargs)
            if transformation_dict is not None:
                self.data_store.set("transformation", transformation_dict)

    def prepare_competitors(self):
        """
        Prepare competitor models already in the preprocessing stage. This is performed here, because some models might
        need to have internet access, which is depending on the operating system not possible during postprocessing.
        This method checks currently only, if the Intelli03-ts-v1 model is requested as competitor and downloads the
        data if required.
        """
        logging.info("Searching for competitors to be prepared for use.")
        competitors = to_list(self.data_store.get_default("competitors", default=[]))
        if len(competitors) > 0:
            for competitor_name in competitors:
                if competitor_name.lower() == "IntelliO3-ts-v1".lower():
                    logging.info("Prepare IntelliO3-ts-v1 model")
                    from mlair.reference_models.reference_model_intellio3_v1 import IntelliO3_ts_v1
                    path = os.path.join(self.data_store.get("competitor_path"), competitor_name)
                    IntelliO3_ts_v1("IntelliO3-ts-v1", path).make_reference_available_locally(remove_tmp_dir=False)
        else:
            logging.info("No preparation required because no competitor was provided to the workflow.")


def f_proc(data_handler, station, name_affix, store, **kwargs):
    """
    Try to create a data handler for given arguments. If build fails, this station does not fulfil all requirements and
    therefore f_proc will return None as indication. On a successfull build, f_proc returns the built data handler and
    the station that was used. This function must be implemented globally to work together with multiprocessing.
    """
    try:
        res = data_handler.build(station, name_affix=name_affix, store_processed_data=store, **kwargs)
    except (AttributeError, EmptyQueryResult, KeyError, requests.ConnectionError, ValueError) as e:
        formatted_lines = traceback.format_exc().splitlines()
        logging.info(
            f"remove station {station} because it raised an error: {e} -> {' | '.join(f_inspect_error(formatted_lines))}")
        res = None
    return res, station


def f_inspect_error(formatted):
    for i in range(len(formatted) - 1, -1, -1):
        if "mlair/mlair" not in formatted[i]:
            return formatted[i - 3:i]
    return formatted[-3:0]
